#include "cstdlib"
#include "common.h"

static UBMP8 K_TR[64] = {
	0, 1, 2, 3, 3, 2, 1, 0,
	1, 4, 5, 6, 6, 5, 4, 1,
	2, 5, 7, 8, 8, 7, 5, 2,
	3, 6, 8, 9, 9, 8, 6, 3,
	3, 6, 8, 9, 9, 8, 6, 3,
	2, 5, 7, 8, 8, 7, 5, 2,
	1, 4, 5, 6, 6, 5, 4, 1,
	0, 1, 2, 3, 3, 2, 1, 0
};
static UBMP8 K1_TR[64] = {
	0, 1, 2, 3, 4, 5, 6, 7,
	1, 8, 9,10,11,12,13,14,
	2, 9,15,16,17,18,19,20,
	3,10,16,21,22,23,24,25,
	4,11,17,22,26,27,28,29,
	5,12,18,23,27,30,31,32,
	6,13,19,24,28,31,33,34,
	7,14,20,25,29,32,34,35
};
static UBMP8 K2_TR[64] = {
	0, 1, 2, 3, 3, 2, 1, 0,
	4, 5, 6, 7, 7, 6, 5, 4,
	8, 9,10,11,11,10, 9, 8,
   12,13,14,15,15,14,13,12,
   16,17,18,19,19,18,17,16,
   20,21,22,23,23,22,21,20,
   24,25,26,27,27,26,25,24,
   28,29,30,31,31,30,29,28,
};

int start_index[] = {
	62, 61, 60, 48, 47, 46
};

static const int rotF = 1;
static const int rotR = 2;
static const int rotD = 4;

static const int piece_order[2][12] =  {
	{bking,wking,bpawn,wpawn,bqueen,wqueen,brook,wrook,bbishop,wbishop,bknight,wknight},
	{wking,bking,wpawn,bpawn,wqueen,bqueen,wrook,brook,wbishop,bbishop,wknight,bknight}
};

static const int piece_cv[15] = {
	0,0,90,50,31,30,10,0,90,50,31,30,10,0
};

static const char piece_name[] = "_kqrbnpkqrbnp_";

BMP16 KK_index[4096];
BMP16 KK_WP_index[4096];
BMP16 KK_rotation[4096];
BMP16 KK_WP_rotation[4096];
BMP16 SS_index[2][3][4096];
BMP32 SSS_index[2][1][64 * 64 * 64];

/*
Initialize index tables and pointers
*/
void init_indices() {
	int temp[2048];
	int index;
	int i,j,k,a,N;
	int u1,u2;
	int rot;

	for( i = 0; i < 4096; i++) {
		KK_index[i] = ILLEGAL;
		KK_WP_index[i] = ILLEGAL;
	}

	/*without pawn*/
	for( i = 0; i < 2048; i++) {
		temp[i] = ILLEGAL;
	}

	index = 0;
	for(i = 0;i < 64;i++) {
		for(j = 0;j < 64;j++) {
			if(distance(SQ6488(i),SQ6488(j)) <= 1)
				continue;
            /*rotations*/
			u1 = i;
			u2 = j;
			rot = 0;

			if(file64(u1) > FILED) {
				u1 = MIRRORF64(u1);
				u2 = MIRRORF64(u2);
				rot ^= rotF;
			}
			if(rank64(u1) > RANK4) {
				u1 = MIRRORR64(u1);
				u2 = MIRRORR64(u2);
				rot ^= rotR;
			}
			if(rank64(u1) > file64(u1)) {
				u1 = MIRRORD64(u1);
				u2 = MIRRORD64(u2);
				rot ^= rotD;
			}
			if(file64(u1) == rank64(u1)) {
				if(rank64(u2) > file64(u2)) {
					u1 = MIRRORD64(u1);
				    u2 = MIRRORD64(u2);
					rot ^= rotD;
				}
			}

            /*actual index*/
			if(file64(u1) == rank64(u1)) {
				u1 = K_TR[u1];
				u2 = K1_TR[u2];
			} else {
				u1 = K_TR[u1];
				u2 = u2;
			}

			if(temp[u1 * 64 + u2] == ILLEGAL) {
				temp[u1 * 64 + u2] = index;
				KK_index[i * 64 + j] = index;
				KK_rotation[i * 64 + j] = rot;
				index++;
			} else {
				KK_index[i * 64 + j] = temp[u1 * 64 + u2];
				KK_rotation[i * 64 + j] = rot;
			}
		}
	}
    /*with pawn*/
	for( i = 0; i < 2048; i++) {
		temp[i] = ILLEGAL;
	}

	index = 0;
	for( i = 0;i < 64;i++) {
		for( j = 0; j < 64;j++) {
			if(distance(SQ6488(i),SQ6488(j)) <= 1)
				continue;

			/*reflection*/
			u1 = i;
			u2 = j;
			rot = 0;
			if(file64(u1) > FILED) {
				u1 = MIRRORF64(u1);
				u2 = MIRRORF64(u2);
				rot ^= rotF;
			}

			/*actual index*/
			u1 = K2_TR[u1];
			u2 = u2;

			if(temp[u1 * 64 + u2] == ILLEGAL) {
				temp[u1 * 64 + u2] = index;
				KK_WP_index[i * 64 + j] = index;
				KK_WP_rotation[i * 64 + j] = rot;
				index++;
			} else {
				KK_WP_index[i * 64 + j] = temp[u1 * 64 + u2];
				KK_WP_rotation[i * 64 + j] = rot;
			}
			/*end*/
		}
	}

	/*clear tables*/
	for(i = 0; i < 3; i++) {
		for(j = 0;j < 4096;j++) {
			SS_index[0][i][j] = ILLEGAL;
			SS_index[1][i][j] = ILLEGAL;
		}
	}
	for(i = 0; i < 1; i++) {
		for(j = 0;j < 262144;j++) {
			SSS_index[0][i][j] = ILLEGAL;
			SSS_index[1][i][j] = ILLEGAL;
		}
	}

	/*generate enumeration tables*/
	int i1,i2;

	for(a = 0;a < 6;a++) {
		if(a >= 3) {
			i1 = 1;
			i2 = a - 3;
		} else {
			i1 = 0;
			i2 = a;
		}
		
		index = 0;
		N = start_index[a];
		for(i = 0;i < N;i++) {
			for(j = i + 1;j < N;j++) {
				SS_index[i1][i2][i * N + j] = index;
				SS_index[i1][i2][j * N + i] = index;
				index++;
			}
		}
	}

	for(a = 0;a < 6;a+=3) {
		if(a >= 3) {
			i1 = 1;
			i2 = 0;
		} else {
			i1 = 0;
			i2 = 0;
		}
		
		index = 0;
		N = start_index[a];
		for(i = 0;i < N;i++) {
			for(j = i + 1;j < N;j++) {
				for(k = j + 1;k < N;k++) {
					SSS_index[i1][i2][i * N * N + j * N + k] = index;
					SSS_index[i1][i2][i * N * N + k * N + j] = index;
					SSS_index[i1][i2][j * N * N + i * N + k] = index;
					SSS_index[i1][i2][j * N * N + k * N + i] = index;
					SSS_index[i1][i2][k * N * N + i * N + j] = index;
					SSS_index[i1][i2][k * N * N + j * N + i] = index;
					index++;
				}
			}
		}
	}
}
/*
initialize enumerator
*/
void ENUMERATOR::init() {
	int t_piece[MAX_PIECES];
	int n_pawns = 0,n_pieces = 0,order = 0,
		i,j,pic;

	/*name*/
	for(i = 0;i < n_piece; i++)
		name[i] = piece_name[piece[i]];
	name[i++] = '.';
	name[i++] = (player == white) ? 'w':'b';
    name[i++] = 0;

	/*determine streaming order*/
	int vcount[2] = {0,0}, stronger;
	for(i = 0;i < n_piece; i++)
		vcount[COLOR(piece[i])] += piece_cv[piece[i]];

	if(vcount[white] > vcount[black]) stronger = white;
	else if(vcount[black] > vcount[white]) stronger = black;
	else stronger = player;

	/*ordered list*/
	for(i = 0;i < n_piece;i++) 
		t_piece[i] = piece[i];

	for(i = 0;i < 12;i++) {
		pic = piece_order[stronger][i];
		for(j = 0;j < n_piece; j++) {
			if(t_piece[j] == pic) {
				if(PIECE(pic) == king) {
					index[order] = 1;
					n_pieces++;
				} else if (PIECE(pic) == pawn) {
					index[order] = 48 - n_pawns;
					n_pawns++;
				} else {
					index[order] = 64 - n_pieces - n_pawns;
					n_pieces++;
				}
				piece[order] = pic;
				order++;
			}
		}
	}
	/*kings*/
	if(n_pawns) index[1] = 1806;
	else index[1] = 462;

	/*same pieces*/
	for(i = 2;i < n_piece; i++) {
		for(j = i + 1;j < n_piece;j++) {
			if(piece[i] != piece[j]) break;
		}
		j--;

		/*for now limit to three similar pieces only*/
        if(j - i > 2) 
			j = i + 2;

		/*calculate number of unique positions*/
		if(i != j) {
			for(int k = i;k < j;k++)  {
				index[j] *= index[k];
				index[k] = 1;
			}
			switch(j - i) {
			case 1: index[j] /= (2 * 1); break;
			case 2: index[j] /= (3 * 2 * 1); break;
			case 3: index[j] /= (4 * 3 * 2 * 1); break;
			case 4: index[j] /= (5 * 4 * 3 * 2 * 1); break;
			}
		}
		i = j;
	}

	/*divisor*/
	divisor[n_piece - 1] = 1;
	for(i = 0;i < n_piece; i++) {
		j = n_piece - 1 - i;
		size *= index[j];
		if(j >= 1) 
			divisor[j - 1] = size;
	}
}
/*
Get index of position
NB: square is modified on return
*/
bool ENUMERATOR::get_index(UBMP32& pindex) {

	/*variables*/
	UBMP32 temp;
	int i,ispawn,N;

	/*illegal pawn placement on kings' square*/
	if(n_pawn) {
		for(i = 2;i < 2 + n_pawn; i++) {
			if(square[i] == square[0] || square[i] == square[1])
				return false;
		}
	}

    /*init index to zero*/
    pindex = 0;

	/*legal placement based on other piece location*/
	for(i = n_piece - 1;i >= 2; i--) {

		/*start and finish*/
		int k,l,start,finish,temp;
		ispawn = (PIECE(piece[i]) == pawn);
		start = (ispawn ? 2 : 0);
		finish = i;
		if(i > 2 && index[i - 1] == 1) {
			finish = i - 1;
			if(i > 3 && index[i - 2] == 1) {
				finish = i - 2;
				if(i > 4 && index[i - 3] == 1) {
					finish = i - 3;
					if(i > 5 && index[i - 4] == 1) {
						finish = i - 4;
					}
				}
			}
		}

		/*legal placement of pieces/pawns*/
		for(k = start; k < finish;k++) 
			t_square[k] = square[k];
		for(k = start; k < finish;k++) {
			for(l = k + 1; l < finish;l++) {
				if(t_square[k] > t_square[l]) {
                    temp = t_square[k];
					t_square[k] = t_square[l];
					t_square[l] = temp;
				}
			}
		}
		
		for(k = finish - 1; k >= start ; k--) {
			if(square[i] >= t_square[k]) square[i]--;
		}
	}

	for(i = n_piece - 1;i >= 2; i--) {
		ispawn = (PIECE(piece[i]) == pawn);
		
		if(index[i - 1] == 1) {
			/*three similar pieces*/
			if(index[i - 2] == 1) {
				if(ispawn) {
					N = start_index[(i - 2) - 2 + 3];
					temp = SSS_index[1][(i - 2) - 2][SQ6448(square[i - 2]) * N * N + SQ6448(square[i - 1]) * N  + SQ6448(square[i])];
			    } else {
					N = start_index[(i - 2) - 2];
					temp = SSS_index[0][(i - 2) - 2][square[i - 2] * N * N + square[i - 1] * N  + square[i]];
				}
				pindex += temp * divisor[i];
				i -= 2;
			/*two  similar pieces*/
			} else {
                if(ispawn) {
					N = start_index[(i - 1) - 2 + 3];
					temp = SS_index[1][(i - 1) - 2][SQ6448(square[i - 1]) * N + SQ6448(square[i])];
				} else {
					N = start_index[(i - 1) - 2];
					temp = SS_index[0][(i - 1) - 2][square[i - 1] * N + square[i]];
				}
				pindex += temp * divisor[i];
				i--;
			}
		} else {
			if(ispawn) temp = SQ6448(square[i]);
			else temp = square[i];	
			pindex += temp * divisor[i];
		}
	}
	/*legal positions of kings*/
	if(n_pawn) temp = KK_WP_index[square[0] * 64 + square[1]];
	else temp = KK_index[square[0] * 64 + square[1]];
	pindex += temp * divisor[1];

	return true;
}
/*
check if the board should be flipped
*/
void ENUMERATOR::check_flip() {

	int i,count[2] = {0,0},vcount[2] = {0,0},pic;
	for(i = 2;i < n_piece; i++) {
		pic = piece[i];
		count[COLOR(pic)]++;
		vcount[COLOR(pic)] += piece_cv[pic];
	}

	if(count[white] > count[black]);
	else if(count[white] == count[black] 
		&& vcount[white] >= vcount[black]);
	else  {
		player = invert(player);
		for(i = 0;i < n_piece;i++) {
			piece[i] = invert_color(piece[i]);
			square[i] = MIRRORR64(square[i]);
		}
	}
}
/*
order pieces based on the streaming order used
by the generator.
*/		
void ENUMERATOR::sort() {
    int t_piece[MAX_PIECES],t_square[MAX_PIECES];
	int i,j,order,pic;
	int vcount[2] = {0,0},stronger;

    /*determine streaming order*/
    for(i = 2;i < n_piece; i++) {
		pic = piece[i];
		vcount[COLOR(pic)] += piece_cv[pic];
	}

    /*who is stronger?*/
	if(vcount[white] > vcount[black]) stronger = white;
	else if(vcount[black] > vcount[white]) stronger = black;
	else stronger = player;

	/*ordered list*/
	for(i = 0;i < n_piece;i++) {
		t_piece[i] = piece[i];
		t_square[i] = square[i];
	}

	order = 0;
	for(i = 0;i < 12;i++) {
		pic = piece_order[stronger][i];
		for(j = 0;j < n_piece; j++) {
			if(t_piece[j] == pic) {
				piece[order] = t_piece[j];
				square[order] = t_square[j];
				order++;
			}
		}
	}
}
/*
Get index
*/
void SEARCHER::get_index(UBMP32& pos_index,UBMP32& tab_index,
			   int player, int w_king, int b_king,
			   int piece1, int square1,
			   int piece2, int square2,
			   int piece3, int square3
			   ) {

	int i,rot,sq;
	ENUMERATOR local_enum,*penum;
	local_enum.clear();
	local_enum.player = player;
	local_enum.add(wking,w_king);
	local_enum.add(bking,b_king);
	local_enum.add(piece1,square1);
	if(piece2) local_enum.add(piece2,square2);
	if(piece3) local_enum.add(piece3,square3);
	local_enum.check_flip();

	tab_index = EGBB::GetIndex(&local_enum);
	penum = &egbbs[tab_index]->enumerator;

	if(local_enum.n_pawn) 
		rot = KK_WP_rotation[local_enum.square[0] * 64 + local_enum.square[1]];
	else 
        rot = KK_rotation[local_enum.square[0] * 64 + local_enum.square[1]];

	for(i = 0;i < local_enum.n_piece;i++) {
		local_enum.divisor[i] = penum->divisor[i];
		local_enum.index[i] = penum->index[i];

		sq = local_enum.square[i];
		if(rot & rotF) sq = MIRRORF64(sq);
		if(rot & rotR) sq = MIRRORR64(sq);
        if(rot & rotD) sq = MIRRORD64(sq);
        local_enum.square[i] = sq;
	}
	local_enum.get_index(pos_index);
}
